from typing import Optional

from spectator.clock import Clock, SystemClock
from spectator.counter import Counter, MonotonicCounter
from spectator.distsummary import DistributionSummary
from spectator.gauge import AgeGauge, Gauge, MaxGauge
from spectator.id import MeterId
from spectator.sidecarconfig import SidecarConfig
from spectator.sidecarwriter import SidecarWriter
from spectator.timer import Timer


class Registry:
    """Registry for reporting data to SpectatorD."""

    def __init__(self, clock: Clock = SystemClock(),
                 config: SidecarConfig = SidecarConfig()) -> None:
        self._clock = clock
        self._common_tags = config.common_tags()
        self._writer = SidecarWriter.create(config.output_location())

    def _merge_common_tags(self, meter_id: MeterId) -> MeterId:
        return meter_id if len(self._common_tags) == 0 else meter_id.with_tags(self._common_tags)

    def _new_meter(self, name: str, tags: Optional[dict] = None) -> MeterId:
        if tags is None:
            tags = {}
        return self._merge_common_tags(MeterId(name, tags))

    def clock(self) -> Clock:
        return self._clock

    def close(self) -> None:
        self._writer.close()

    def age_gauge(self, name: str, tags: Optional[dict] = None) -> AgeGauge:
        return AgeGauge(self._new_meter(name, tags), writer=self._writer)

    def counter(self, name: str, tags: Optional[dict] = None) -> Counter:
        return Counter(self._new_meter(name, tags), writer=self._writer)

    def distribution_summary(self, name: str, tags: Optional[dict] = None) -> DistributionSummary:
        return DistributionSummary(self._new_meter(name, tags), writer=self._writer)

    def pct_distribution_summary(self, name: str,
                                 tags: Optional[dict] = None) -> DistributionSummary:
        return DistributionSummary(self._new_meter(name, tags), meter_type="D", writer=self._writer)

    def gauge(self, name: str, tags: Optional[dict] = None) -> Gauge:
        return Gauge(self._new_meter(name, tags), writer=self._writer)

    def max_gauge(self, name: str, tags: Optional[dict] = None) -> MaxGauge:
        return MaxGauge(self._new_meter(name, tags), writer=self._writer)

    def monotonic_counter(self, name: str, tags: Optional[dict] = None) -> MonotonicCounter:
        return MonotonicCounter(self._new_meter(name, tags), writer=self._writer)

    def timer(self, name: str, tags: Optional[dict] = None) -> Timer:
        return Timer(self._new_meter(name, tags), clock=self._clock, writer=self._writer)

    def pct_timer(self, name: str, tags: Optional[dict] = None) -> Timer:
        return Timer(self._new_meter(name, tags), clock=self._clock, meter_type="T",
                     writer=self._writer)

    def start(self, config: Optional[dict] = None) -> None:
        """Avoid breaking the API."""

    def clear_meters_and_start(self) -> None:
        """Avoid breaking the API."""

    def stop(self) -> None:
        """Avoid breaking the API."""

    def stop_without_publish(self) -> None:
        """Avoid breaking the API."""

    def __iter__(self) -> "RegistryIterator":
        """Avoid breaking the API."""
        return RegistryIterator([])


class RegistryIterator:
    """Avoid breaking the API."""

    def __init__(self, meters):
        pass

    def __next__(self):
        raise StopIteration



import time


class Clock:

    def wall_time(self) -> float:
        raise NotImplementedError("use a concrete implementation")

    def monotonic_time(self) -> float:
        raise NotImplementedError("use a concrete implementation")


class SystemClock(Clock):

    def wall_time(self) -> float:
        return time.time()

    def monotonic_time(self) -> float:
        """The time.perf_counter() function was added in Python 3.3. We expect Python >= 3.5."""
        return time.perf_counter()


class ManualClock(Clock):

    def __init__(self, wall_init: float = 0, monotonic_init: float = 0) -> None:
        self._wall = wall_init
        self._monotonic = monotonic_init

    def wall_time(self) -> float:
        return self._wall

    def monotonic_time(self) -> float:
        return self._monotonic

    def set_wall_time(self, t: float) -> None:
        self._wall = t

    def set_monotonic_time(self, t: float) -> None:
        self._monotonic = t



from typing import Optional

from spectator.id import MeterId
from spectator.sidecarmeter import SidecarMeter
from spectator.sidecarwriter import SidecarWriter


class Gauge(SidecarMeter):
    """The value is a number that was sampled at a point in time. The default time-to-live (TTL)
    for gauges is 900 seconds (15 minutes) - they will continue reporting the last value set for
    this duration of time. An optional ttl_seconds may be set to control the lifespan of these
    values. SpectatorD enforces a minimum ttl of 5 seconds."""

    def __init__(self, meter_id: MeterId, ttl_seconds: Optional[int] = None,
                 writer: SidecarWriter = SidecarWriter.create("none")) -> None:
        if ttl_seconds is None:
            meter_type = "g"
        else:
            meter_type = "g,{}".format(ttl_seconds)
        super().__init__(meter_id, meter_type)
        self._writer = writer

    def set(self, value: float) -> None:
        self._writer.write(self.idString, value)

    @staticmethod
    def get() -> float:
        """Avoid breaking the API."""
        return 0


class AgeGauge(SidecarMeter):
    """The value is the time in seconds since the epoch at which an event has successfully
    occurred, or 0 to use the current time in epoch seconds. After an Age Gauge has been set,
    it will continue reporting the number of seconds since the last time recorded, for as long
    as the SpectatorD process runs. This meter type makes it easy to implement the Time Since
    Last Success alerting pattern."""

    def __init__(self, meter_id: MeterId,
                 writer: SidecarWriter = SidecarWriter.create("none")) -> None:
        super().__init__(meter_id, "A")
        self._writer = writer

    def set(self, seconds: int) -> None:
        self._writer.write(self.idString, seconds)


class MaxGauge(SidecarMeter):
    """The value is a number that was sampled at a point in time, but it is reported as a maximum
    gauge value to the backend."""

    def __init__(self, meter_id: MeterId,
                 writer: SidecarWriter = SidecarWriter.create("none")) -> None:
        super().__init__(meter_id, "m")
        self._writer = writer

    def set(self, value: float) -> None:
        self._writer.write(self.idString, value)



import re

from spectator.id import MeterId


class SidecarMeter:
    ALLOWED_CHARS = re.compile("[^-._A-Za-z0-9~^]")

    def __init__(self, meter_id: MeterId, meter_type: str):
        """
        :param meter_id:
            Base identifier for all measurements supplied by this meter.
        :param meter_type:
            Prefix string for line to output to SpectatorD.
        """
        self.meterId = meter_id
        self.idString = self._create_id_string(meter_id, meter_type)

    def _replace_invalid_chars(self, s: str) -> str:
        return self.ALLOWED_CHARS.sub("_", s)

    def _create_id_string(self, meter_id: MeterId, meter_type: str) -> str:
        s = "{}:{}".format(meter_type, self._replace_invalid_chars(meter_id.name))

        for key, value in sorted(meter_id.tags().items()):
            k = self._replace_invalid_chars(key)
            v = self._replace_invalid_chars(value)
            s += ",{}={}".format(k, v)

        return s + ":"



from spectator.clock import Clock, SystemClock
from spectator.id import MeterId
from spectator.sidecarmeter import SidecarMeter
from spectator.sidecarwriter import SidecarWriter


class Timer(SidecarMeter):
    """The value is the number of seconds that have elapsed for an event. A stopwatch method
    is available, which provides a context manager that can be used to automate recording the
    timing for a block of code using the `with` statement.

    Two types are available: (1) standard timers, with a meter type of "t", and (2) percentile
    timers, with a meter type of "T". The standard timers are the default.

    In order to maintain the data distribution, Percentile Timers have a higher storage cost,
    with a worst-case of up to 300X that of a standard Timer. Be diligent about any additional
    dimensions added to Percentile Timers and ensure that they have a small bounded cardinality."""

    def __init__(self, meter_id: MeterId, clock: Clock = SystemClock(), meter_type: str = "t",
                 writer: SidecarWriter = SidecarWriter.create("none")) -> None:
        if meter_type not in ["t", "T"]:
            raise ValueError("Timers must have a meter type of 't' or 'T'.")

        super().__init__(meter_id, meter_type)
        self._clock = clock
        self._writer = writer

    def record(self, seconds: float) -> None:
        if seconds >= 0:
            self._writer.write(self.idString, seconds)

    def stopwatch(self) -> "StopWatch":
        return StopWatch(self)

    def clock(self) -> Clock:
        return self._clock

    @staticmethod
    def count() -> int:
        """Avoid breaking the API."""
        return 0

    @staticmethod
    def total_time() -> float:
        """Avoid breaking the API."""
        return 0


class StopWatch:
    """Context Manager that records the duration of a with-statement block in a Timer or
    PercentileTimer. The type annotation for the init parameter is skipped to avoid circular
    imports."""

    def __init__(self, timer) -> None:
        self._timer = timer

    def __enter__(self) -> None:
        self._start = self._timer.clock().monotonic_time()

    def __exit__(self, type_, value, traceback) -> None:
        self._timer.record(self.duration())

    def duration(self) -> float:
        now = self._timer.clock().monotonic_time()
        return now - self._start



from typing import Dict, Any

from spectator.commontags import common_tags


class SidecarConfig:
    """To provide a custom configuration for Spectator-py, pass a dictionary to the constructor
    which returns values for the keys listed below. Omitting keys will cause this class to return
    the default values.

    {
      "sidecar.common-tags": {"key": "value"},
      "sidecar.output-location": "location"
    }

    Then provide the custom instance to a new instance of the Registry, rather than relying on
    the GlobalRegistry, which has a default configuration. For example:

    r = Registry(config=SidecarConfig({"sidecar.output-location": "memory"}))
    r.counter("test").increment()
    """

    def __init__(self, config: Dict[str, Any] = None) -> None:
        """Optionally initialize with a dictionary which returns values that override the
        defaults."""
        if config is None:
            self._config = {}
        else:
            self._config = config

    def common_tags(self) -> Dict[str, str]:
        """Common infrastructure tags from the Netflix environment variables."""
        return self._config.get("sidecar.common-tags", common_tags())

    def output_location(self) -> str:
        """
        The location where data will be emitted. Supported values include:

          * `none` - disable output
          * `stdout` - write to standard out for the process
          * `stderr` - write to standard error for the process
          * `file://$path_to_file` - write to a file (e.g. file:///tmp/foo/bar)
          * `udp://$host:$port` - write to a UDP socket

        The default value is the port used by SpectatorD.
        """
        return self._config.get("sidecar.output-location", "udp://127.0.0.1:1234")



from spectator import Clock
from spectator.timer import StopWatch
from spectator.registry import Registry

from typing import Dict, Optional


class PercentileBuckets:
    """Avoid breaking the API."""

    @staticmethod
    def get(i) -> int:
        return 0

    @staticmethod
    def length() -> int:
        return 0

    @staticmethod
    def index_of(v) -> int:
        return 0

    @staticmethod
    def bucket(v) -> int:
        return 0

    @staticmethod
    def percentiles(counts, pcts, results) -> None:
        pass

    @staticmethod
    def percentile(counts, p) -> float:
        return 0


class PercentileTimer:
    """
    This class is deprecated, but it is retained to avoid breaking the API.

    Uses of this class should be replaced with direct use of the Registry helper method:

    from spectator import GlobalRegistry

    GlobalRegistry.pct_timer("server.requestLatency").record(0.01)
    """

    def __init__(self, registry: Registry, name: str, tags: Optional[Dict[str, str]] = None,
                 min: float = 0, max: float = 0) -> None:
        if tags is None:
            tags = {}
        self._clock = registry.clock()
        self._pct_timer = registry.pct_timer(name, tags)

    def record(self, amount: float) -> None:
        self._pct_timer.record(amount)

    def stopwatch(self) -> StopWatch:
        return StopWatch(self)

    def clock(self) -> Clock:
        return self._clock

    @staticmethod
    def count() -> int:
        return 0

    @staticmethod
    def total_time() -> float:
        return 0

    @staticmethod
    def percentile(p) -> float:
        return 0


class PercentileDistributionSummary:
    """
    This class is deprecated, but it is retained to avoid breaking the API.

    Uses of this class should be replaced with direct use of the Registry helper method:

    from spectator import GlobalRegistry

    GlobalRegistry.pct_distribution_summary("server.requestSize").record(10)
    """

    def __init__(self, registry: Registry, name: str, tags: Optional[Dict[str, str]] = None,
                 min: int = 0, max: int = 0) -> None:
        if tags is None:
            tags = {}
        self._pct_distsummary = registry.pct_distribution_summary(name, tags)

    def record(self, amount) -> None:
        self._pct_distsummary.record(amount)

    @staticmethod
    def count() -> int:
        return 0

    @staticmethod
    def total_amount() -> int:
        return 0

    @staticmethod
    def percentile(p) -> int:
        return 0



import threading


class AtomicNumber:

    def __init__(self, init: float) -> None:
        self._value = init
        self._lock = threading.RLock()

    def set(self, v: float) -> None:
        with self._lock:
            self._value = v

    def get(self) -> float:
        with self._lock:
            return self._value

    def get_and_set(self, v: float) -> float:
        with self._lock:
            tmp = self._value
            self._value = v
            return tmp

    def get_and_increment(self) -> float:
        return self.get_and_add(1)

    def increment_and_get(self) -> float:
        return self.add_and_get(1)

    def get_and_add(self, amount: float) -> float:
        with self._lock:
            tmp = self._value
            self._value += amount
            return tmp

    def add_and_get(self, amount: float) -> float:
        with self._lock:
            self._value += amount
            return self._value

    def max(self, amount: float) -> float:
        with self._lock:
            self._value = max(self._value, amount)
            return self._value

    def __str__(self) -> str:
        with self._lock:
            return "AtomicLong({})".format(self._value)



import os

from typing import Dict


def _add_non_empty(tags: Dict[str, str], tag: str, *env_vars: str) -> None:
    for env_var in env_vars:
        value = os.environ.get(env_var)
        if value is None:
            continue
        value = value.strip()
        if len(value) != 0:
            tags[tag] = value
            break


def common_tags() -> Dict[str, str]:
    """Extract common infrastructure tags from the Netflix environment variables, which are
    specific to a process and thus cannot be managed by a shared SpectatorD instance."""
    tags = {}
    _add_non_empty(tags, "nf.container", "TITUS_CONTAINER_NAME")
    _add_non_empty(tags, "nf.process", "NETFLIX_PROCESS_NAME")
    return tags



from spectator.id import MeterId
from spectator.sidecarmeter import SidecarMeter
from spectator.sidecarwriter import SidecarWriter


class DistributionSummary(SidecarMeter):
    """The value tracks the distribution of events. It is similar to a Timer, but more general,
    because the size does not have to be a period of time. For example, it can be used to
    measure the payload sizes of requests hitting a server or the number of records returned
    from a query.

    Two types are available: (1) standard distribution summaries, with a meter type of "d", and (2)
    percentile distribution summaries, with a meter type of "D". The standard distribution summaries
    are the default.

    In order to maintain the data distribution, Percentile Distribution Summaries have a higher
    storage cost, with a worst-case of up to 300X that of a standard Distribution Summary. Be
    diligent about any additional dimensions added to Percentile Distribution Summaries and ensure
    that they have a small bounded cardinality."""

    def __init__(self, meter_id: MeterId, meter_type: str = "d",
                 writer: SidecarWriter = SidecarWriter.create("none")) -> None:
        if meter_type not in ["d", "D"]:
            raise ValueError("Distribution Summaries must have a meter type of 'd' or 'D'.")
        super().__init__(meter_id, meter_type)
        self._writer = writer

    def record(self, amount: int) -> None:
        if amount >= 0:
            self._writer.write(self.idString, amount)

    @staticmethod
    def count() -> int:
        """Avoid breaking the API."""
        return 0

    @staticmethod
    def total_amount() -> int:
        """Avoid breaking the API."""
        return 0



from typing import Dict, Optional


class MeterId:
    def __init__(self, name: str, tags: Optional[Dict[str, str]] = None) -> None:
        if tags is None:
            self._tags = {}
        else:
            self._tags = tags
        self.name = name

    def tags(self) -> Dict[str, str]:
        return self._tags.copy()

    def with_stat(self, v: str) -> "MeterId":
        """Avoid breaking the API."""
        return self

    def with_default_stat(self, v: str) -> "MeterId":
        """Avoid breaking the API."""
        return self

    def with_tag(self, k: str, v: str) -> "MeterId":
        tags = self._tags.copy()
        tags[k] = v
        return MeterId(self.name, tags)

    def with_tags(self, new_tags: Dict[str, str]) -> "MeterId":
        tags = self._tags.copy()
        tags.update(new_tags)
        return MeterId(self.name, tags)

    def __hash__(self):
        return hash((self.name, frozenset(self._tags.items())))

    def __eq__(self, other) -> bool:
        if isinstance(self, other.__class__):
            return self.__dict__ == other.__dict__
        return False

    def __str__(self) -> str:
        if len(self._tags) > 0:
            tags = ["{}={}".format(k, v) for k, v in sorted(self._tags.items())]
            return "{},{}".format(self.name, ",".join(tags))
        else:
            return self.name



from spectator.id import MeterId
from spectator.sidecarmeter import SidecarMeter
from spectator.sidecarwriter import SidecarWriter


class Counter(SidecarMeter):
    """The value is the number of increments that have occurred since the last time it was
    recorded. The value will be reported to the Atlas backend as a rate-per-second."""

    def __init__(self, meter_id: MeterId,
                 writer: SidecarWriter = SidecarWriter.create("none")) -> None:
        super().__init__(meter_id, "c")
        self._writer = writer

    def increment(self, delta: int = 1) -> None:
        if delta > 0:
            self._writer.write(self.idString, delta)

    @staticmethod
    def count() -> int:
        """Avoid breaking the API."""
        return 0


class MonotonicCounter(SidecarMeter):
    """The value is a monotonically increasing number. A minimum of two samples must be received
    in order for SpectatorD to calculate a delta value and report it to the backend."""

    def __init__(self, meter_id: MeterId,
                 writer: SidecarWriter = SidecarWriter.create("none")) -> None:
        super().__init__(meter_id, "C")
        self._writer = writer

    def set(self, amount: int) -> None:
        self._writer.write(self.idString, amount)



import os

from setuptools import setup


def read(fname):
    """Utility function to read a file, for publishing the README with the package."""
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="netflix-spectator-py",
    version="0.2.0",
    python_requires=">3.5",
    description="Python library for reporting metrics to the Netflix Atlas Timeseries Database.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="Netflix Telemetry Engineering",
    author_email="netflix-atlas@googlegroups.com",
    license="Apache 2.0",
    url="https://github.com/Netflix/spectator-py",
    packages=["spectator", "spectator.histogram"],
    install_requires=[],
    extras_require={
        "dev": [
            "check-manifest",
            "pylint",
            "pytest-cov",
            "pytest"
        ]
    }
)



import logging
import socket
import sys
from typing import TextIO, Tuple, Union
from urllib.parse import urlparse


class SidecarWriter:
    """Base type for a writer that accepts the SpectatorD line protocol. The concrete
    implementations must exist in this module to avoid circular imports."""

    def __init__(self, location: str):
        self._location = location
        self._logger = logging.getLogger("spectator.SidecarWriter")
        self._debug = None

    @staticmethod
    def create(location: str) -> "SidecarWriter":
        """Create a new writer based on a location string."""
        if location == "none":
            return NoopWriter()
        elif location == "memory":
            return MemoryWriter()
        elif location == "stderr":
            return PrintWriter(location, sys.stderr)
        elif location == "stdout":
            return PrintWriter(location, sys.stdout)
        elif location.startswith("file://"):
            # expects: file:///path/to/file
            file = open(urlparse(location).path, "a", encoding="utf-8")
            return PrintWriter(location, file)
        elif location.startswith("udp://"):
            parsed = urlparse(location)
            address = (parsed.hostname, int(parsed.port))
            return UdpWriter(location, address)
        else:
            raise ValueError("unsupported location: {}".format(location))

    def write_impl(self, line: str) -> None:
        """Custom writers should override this method."""
        raise NotImplementedError("use a concrete implementation")

    def write_line(self, line: str) -> None:
        if self._debug is None:
            # on the first write, cache the log level to speed up later comparisons
            self._debug = self._logger.isEnabledFor(logging.DEBUG)
        try:
            if self._debug:
                # when enabled, this log line reduces udp socket performance by 10%
                self._logger.debug("writing to %s: %s", self._location, line)
            self.write_impl(line)
        except IOError:
            self._logger.warning("write to %s failed: %s", self._location, line)

    def write(self, prefix: str, value: Union[int, float]) -> None:
        self.write_line(prefix + str(value))

    def close(self) -> None:
        """Custom writers should override this method."""
        raise NotImplementedError("use a concrete implementation")


class MemoryWriter(SidecarWriter):
    """Writer that stores data in a list, to support unit testing."""

    def __init__(self) -> None:
        super().__init__("memory")
        self._messages = []

    def write_impl(self, line: str) -> None:
        self._messages.append(line)

    def close(self) -> None:
        self._messages.clear()

    def is_empty(self) -> bool:
        return len(self._messages) == 0

    def last_line(self) -> str:
        return self._messages[-1]


class NoopWriter(SidecarWriter):
    """Writer that does nothing. Used to disable output."""

    def __init__(self):
        super().__init__("none")

    def write_impl(self, line: str) -> None:
        pass

    def close(self) -> None:
        pass


class PrintWriter(SidecarWriter):
    """Writer that outputs data to a TextIO instance."""

    def __init__(self, location: str, file: TextIO) -> None:
        super().__init__(location)
        self._file = file

    def write_impl(self, line: str) -> None:
        print(line, file=self._file)

    def close(self) -> None:
        self._file.close()


class UdpWriter(SidecarWriter):
    """Writer that outputs data to UDP socket."""

    def __init__(self, location: str, address: Tuple[str, int]) -> None:
        super().__init__(location)
        self._address = address
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._sock.connect(self._address)

    def write_impl(self, line: str) -> None:
        self._sock.send(bytes(line, encoding="utf-8"))

    def close(self) -> None:
        self._sock.close()
